/**********************************************************************
  Copyright (c) 2007-2008, Space Machines Laboratory, Tohoku University
  Copyright (c) 2007-2008, PieCake,Inc.
**********************************************************************/

#ifndef __ENGINE_REDUNDANCY_CPP__
#define __ENGINE_REDUNDANCY_CPP__


#include "EngineRedundancy.h"



//#define MAX_IK_ERROR 1.0E-10
//#define MAX_IK_ERROR 1.0E-8
#define MAX_IK_ERROR 1.0E-6



#define CHECK_INITIALIZED(return_value)         \
  if (!isInit()) {                              \
    return return_value;                        \
  }




dmatrix calcAdjMatrix(const dmatrix &J) //余因子行列を求める関数
{
  dmatrix Result(J.size1(),J.size2());
  dmatrix J_ij(J.size1()-1,J.size2()-1);

  double  det_J_ij;

  for(unsigned int i=0;i<J.size1();i++){
    for(unsigned int j=0;j<J.size2();j++){

      for(unsigned int m=0;m<J.size1()-1;m++){
	for(unsigned int n=0;n<J.size2()-1;n++){
	  int a=m; 
	  if(m>=i) a=m+1;

          int b=n;
	  if(n>=j) b=n+1;

          J_ij(m,n)=J(a,b);
	}
      }
 
      det_J_ij = det(J_ij);

      if((i+j)%2){
	Result(j,i) = -det_J_ij;
      }
      else{
	Result(j,i) = det_J_ij;
      }

    }
  }

  return Result;
}




EngineRedundancy::EngineRedundancy()
  : is_init(false)
{
  condition_number = 0.0;
}



EngineRedundancy::~EngineRedundancy()
{
  finish();
}






////////////////////////////////////////////////////////////////////
//
//    public function
//
////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------
//  initialize
//------------------------------------------------------------------

bool EngineRedundancy::init(JointName* in_jointname)
{
  ////////  set joint name  ////////
  jointname = in_jointname;

  if(jointname == NULL) {
    cout << "error : [Engine] [init] : set JointName" << endl;
    return false;
  }

  if(!jointname->isInit()) {
    cout << "error : [Engine] [init] : not set configuration" << endl;
    return false;
  }


  
  ////////  create standard model  ////////
  // current model
  cur_model = new StandardModel();
  if( !createRealModelFromJointNameMap(cur_model->body, jointname) ) {
    cout << "error : [Engine] [init] : can not create current model (real)" << endl;
    return false;
  }
  else {
    if( !createStandardModelFromJointNameMap(cur_model, jointname) ) {
      cout << "error : [Engine] [init] : can not create current model (standard)" << endl;
      return false;
    }
  }

  // reference model
  ref_model = new StandardModel();
  if( !createRealModelFromJointNameMap(ref_model->body, jointname) ) {
    cout << "error : [Engine] [init] : can not create reference model (real)" << endl;
    return false;
  }
  else {
    if( !createStandardModelFromJointNameMap(ref_model, jointname) ) {
      cout << "error : [Engine] [init] : can not create reference model (standard)" << endl;
      return false;
    }
  }



  ////////  reference data  ////////
  for(int i = 0; i < 2; i++) {
    dst_pos[i] = vector3(0, 0, 0);
    dst_rot[i] = tvmet::identity<matrix33>();
  }
  dst_redun      = dzerovector(8);




  ////////  others  ////////
  use_move  = false;
  use_waist = false;
  select_move  = dzeromatrix(3,3);
  select_waist = dzeromatrix(3,3);
  if(ref_model->part("MOVE")->numRealJoints() > 0) {
    use_move    = true;
    select_move = didentity(3);
  }
  if(ref_model->part("WAIST")->numRealJoints() > 0) {
    use_waist    = true;
    select_waist = didentity(3);
  }
  use_elbow_r = true;
  use_elbow_l = true;



  //////// initialize finish  ////////
  is_init = true;
  cout << "------- engine init finish -------------------------" << endl;


  return true;
}



void EngineRedundancy::finish()
{
  jointname = NULL;

  delete cur_model;
  delete ref_model;

  is_init = false;
}






//------------------------------------------------------------------
//  main calc function
//------------------------------------------------------------------

bool EngineRedundancy::update(short mode, RightLeft type, CORBA::Boolean isHoldOtherArm)
{
  CHECK_INITIALIZED(false);


  // set reference model to current model
  for(int i = 0; i < cur_model->numVirtualJoints(); i++) {
    ref_model->virtualJoint(i)->q( cur_model->virtualJoint(i)->q() );
  }
  ref_model->calcForwardKinematics();



  bool calcSuccess = true;  
  if(mode == 0) {
    switch(type) {
      
    case RIGHT: {
      //----------------------------------------------------------------------------
      // CP motion 1 or 2
      //----------------------------------------------------------------------------
      if(use_elbow_r) {
	if( !calcMinimizeNorm(dst_pos[RIGHT], dst_rot[RIGHT], RIGHT, use_waist, use_move) )
	  calcSuccess = false;
      }
      else {
	if( !calcElbowAngle(dst_pos[RIGHT], dst_rot[RIGHT], dst_redun[6], RIGHT) )
	  calcSuccess = false;
      }

      //----------------------------------------------------------------------------
      // HOLD motion 1
      //----------------------------------------------------------------------------
      if(isHoldOtherArm) {

	if(use_elbow_l) {
	  if( !calcMinimizeNorm(dst_pos[LEFT], dst_rot[LEFT], LEFT) )
	    calcSuccess = false;
	}
	else {
	  if( !calcElbowAngle(dst_pos[LEFT], dst_rot[LEFT], dst_redun[7], LEFT) )
	    calcSuccess = false;
	}
      }
      break;
    } // end case of RIGHT
      

    case LEFT: {
      //----------------------------------------------------------------------------
      // CP motion 1 or 2
      //----------------------------------------------------------------------------
      if(use_elbow_l) {
	if( !calcMinimizeNorm(dst_pos[LEFT], dst_rot[LEFT], LEFT, use_waist, use_move) )
	  calcSuccess = false;
      }
      else {
	if( !calcElbowAngle(dst_pos[LEFT], dst_rot[LEFT], dst_redun[7], LEFT) )
	  calcSuccess = false;
      }

      //----------------------------------------------------------------------------
      // HOLD motion 1
      //----------------------------------------------------------------------------
      if(isHoldOtherArm) {

	if(use_elbow_r) {
	  if( !calcMinimizeNorm(dst_pos[RIGHT], dst_rot[RIGHT], RIGHT) )
	    calcSuccess = false;
	}
	else {
	  if( !calcElbowAngle(dst_pos[RIGHT], dst_rot[RIGHT], dst_redun[6], RIGHT) )
	    calcSuccess = false;
	}
      }
      break;
    } // end case of LEFT
      

    case DUAL: {
      if(use_elbow_r && use_elbow_l) {
	if( !calcMinimizeNormDual(dst_pos[RIGHT], dst_rot[RIGHT], dst_pos[LEFT], dst_rot[LEFT], use_waist, use_move) )
	  calcSuccess = false;
      }
      else if(!use_elbow_r && !use_elbow_l) {
	if( !calcElbowAngle(dst_pos[RIGHT], dst_rot[RIGHT], dst_redun[6], RIGHT) )
	  calcSuccess = false;
      
	if( !calcElbowAngle(dst_pos[LEFT], dst_rot[LEFT], dst_redun[7], LEFT) )
	  calcSuccess = false;
      }
      else if(use_elbow_r) {
	if( !calcMinimizeNorm(dst_pos[RIGHT], dst_rot[RIGHT], RIGHT, use_waist, use_move) )
	  calcSuccess = false;
	
	if( !calcElbowAngle(dst_pos[LEFT], dst_rot[LEFT], dst_redun[7], LEFT) )
	  calcSuccess = false;
      }
      else {
	if( !calcMinimizeNorm(dst_pos[LEFT], dst_rot[LEFT], LEFT, use_waist, use_move) )
	  calcSuccess = false;

	if( !calcElbowAngle(dst_pos[RIGHT], dst_rot[RIGHT], dst_redun[6], RIGHT) )
	  calcSuccess = false;
      }

    } // end of case DUAL
      
    } // end of switch

  } // end of if (mode = 0)
  //
  //
  //
  //----------------------------------------------------------------------------
  // HOLD motion 2  &&  CP + redun
  //----------------------------------------------------------------------------
  else if(mode == 1 || mode == 2) {

    ////////  move & waist  //////// 
    for(int i = 0; i < 6; i++) {
      ref_model->virtualJoint(i)->q(dst_redun[i]);
    }
    ref_model->calcForwardKinematics();



    ////////  arm  //////// 
    switch(type) {

    case RIGHT: {
      if( !calcElbowAngle(dst_pos[RIGHT], dst_rot[RIGHT], dst_redun[6], RIGHT) )
	calcSuccess = false;
      break;
    } // end of case RIGHT


    case LEFT: {
      if( !calcElbowAngle(dst_pos[LEFT], dst_rot[LEFT], dst_redun[7], LEFT) )
	calcSuccess = false;
      break;
    } // end of case LEFT


    case DUAL: {
      if( !calcElbowAngle(dst_pos[RIGHT], dst_rot[RIGHT], dst_redun[6], RIGHT) )
	calcSuccess = false;
      
      if( !calcElbowAngle(dst_pos[LEFT], dst_rot[LEFT], dst_redun[7], LEFT) )
	calcSuccess = false;
      
    } // end of case DUAL
      
    } // end of switch

  } // end of if(mode == 1)



  ////////  inverse kinematics failuer  ////////
  if(!calcSuccess) {
    // for debug
    checkResults(DUAL, true, true);

    
    // 計算に失敗したら現在値に戻すべきかどうか...CPとPTPで閾値が変わってくる...収束計算にすべきかどうか...
    // set reference model to current model
    for(int i = 0; i < cur_model->numVirtualJoints(); i++) {
      ref_model->virtualJoint(i)->q( cur_model->virtualJoint(i)->q() );
    }
    ref_model->calcForwardKinematics();
    
  }
 

  return calcSuccess;
}



void EngineRedundancy::setProperty(const double dt, const double limitPerSec)
{
  limmiter = dt * limitPerSec;
}






//------------------------------------------------------------------
//  for current data
//------------------------------------------------------------------

bool EngineRedundancy::setCurAngle(const TimedDoubleSeq q)
{
  CHECK_INITIALIZED(false);


  ////////  check data size  ////////
  if ((int)q.data.length() > cur_model->numVirtualJoints()) {
    cout << "error : [Engine] [setCurAngle] : Not much data size" << endl;
    return false;
  }


  ////////  set virtual joint angle  ////////
  for(unsigned int i = 0; i < q.data.length(); i++) {
    cur_model->virtualJoint(i)->q(q.data[i]);
  }


  ////////  update current model  ////////
  cur_model->calcForwardKinematics();


  return true;
}



TimedDoubleSeq EngineRedundancy::getCurPosition()
{
  CHECK_INITIALIZED(TimedDoubleSeq());
  
  
  TimedDoubleSeq xyzR;
  xyzR.data.length(24);
  
  string end_joint_name[] = {"R_END", "L_END"};
  
  for(int i = 0; i < 2; i++) {

    Link* end_link = cur_model->virtualJoint(end_joint_name[i])->endLink();
   
    vector3  p(end_link->p);
    matrix33 R(end_link->R * end_link->Rs);
    
    xyzR.data[0 + 12*i] = p[0];  // x
    xyzR.data[1 + 12*i] = p[1];  // y
    xyzR.data[2 + 12*i] = p[2];  // z
    
    xyzR.data[3 + 12*i]  = R(0,0);
    xyzR.data[4 + 12*i]  = R(0,1);
    xyzR.data[5 + 12*i]  = R(0,2);
    xyzR.data[6 + 12*i]  = R(1,0);
    xyzR.data[7 + 12*i]  = R(1,1);
    xyzR.data[8 + 12*i]  = R(1,2);
    xyzR.data[9 + 12*i]  = R(2,0);
    xyzR.data[10 + 12*i] = R(2,1);
    xyzR.data[11 + 12*i] = R(2,2);
    
  }
  
  return xyzR;
}



TimedDoubleSeq EngineRedundancy::getCurRedundant()
{
  CHECK_INITIALIZED(TimedDoubleSeq());

  
  TimedDoubleSeq phi;
  phi.data.length(8);

  
  for(int i = 0; i < 8; i++)
    phi.data[i] = 0.0;
  
  
  // move & waist
  for (unsigned int i = 0; i < 6; ++i) {
    phi.data[i] = cur_model->virtualJoint(i)->q();
  }
  
  
  // elbow
  phi.data[6] = getCurElbowAngle(RIGHT);
  phi.data[7] = getCurElbowAngle(LEFT);
  
  
  return phi;
}


double EngineRedundancy::getCurElbowAngle(RightLeft type)
{
  CHECK_INITIALIZED(double ());
  

  StandardModelPart* part_arm;
  Link* wrist;
  if(type == RIGHT) {
    part_arm = cur_model->part("RARM");
    wrist  = cur_model->virtualJoint("R_END")->endLink(); 
  }
  else if(type == LEFT) {
    part_arm = cur_model->part("LARM");
    wrist  = cur_model->virtualJoint("L_END")->endLink(); 
  }
  else {
    cout << "error : [Engine] [getCurElbowAngle] : No such arm" << endl;
    return 0;
  }
  
  
  if( part_arm->virtualJoint(3)->empty() ) {
    return 0;
  }

  
  if( part_arm->numRealJoints() < 7 ) {
    return 0;
  }
  
  
  // 肘角度の計算
  Link* shoulder = part_arm->rootVirtualJoint()->rootLink();
  Link* elbow    = part_arm->virtualJoint(3)->endLink();
 


  vector3 w(wrist->p - shoulder->p);   // 肩->手首ベクトル
  vector3 hw(tvmet::normalize(w));     // wのnormalized vector(hはhatの意味)
  vector3 e(elbow->p - shoulder->p);   // 肩->肘ベクトル
  vector3 hv(0, 0, 1);                 // 基準面作成用ベクトル
  vector3 d(hw * (tvmet::dot(hw, e))); // wが作る直線と肘からおろした垂線上にある点
  vector3 p(e - d);                    // 点dから肘へのベクトル (p = e - d)


  double at0,at1;
  at0 = tvmet::dot(hw, tvmet::cross(hv, p));
  at1 = tvmet::dot(hv, p);
  double phi = atan2(at0, at1);  // 肘角度


  return phi;
}




//------------------------------------------------------------------
//  for reference data
//------------------------------------------------------------------

bool EngineRedundancy::setRefPosition(const TimedDoubleSeq p)
{
  CHECK_INITIALIZED(false);


  ////////  check data size  ////////
  if (p.data.length() != 24) {
    cout << "error : [Engine] [setRefPosition] : Not much data size" << endl;
    return false;
  }
  
  
  ////////  set reference data  ////////
  for(int i = 0; i < 2; i++) {
    int l = 3;
    for(int j = 0; j < 3; j++) {
      dst_pos[i](j) = p.data[j + i*12];

      for(int k = 0; k < 3; k++) {
	dst_rot[i](j,k) = p.data[l + i*12];
	l++;
      }
    }
  }

  
  return true;
}



bool EngineRedundancy::setRefRedundant(const TimedDoubleSeq phi)
{
  CHECK_INITIALIZED(false);
  

  ////////  check data size  ////////
  if (phi.data.length() != dst_redun.size()) {
    cout << "error : [Engine] [setRefRedundant] : Not much data size" << endl;
    return false;
  }
  
  
  ////////  set reference data  ////////
  for(unsigned int i = 0; i < phi.data.length(); i++)
    dst_redun(i) = phi.data[i];
  
  
  return true;
}



TimedDoubleSeq EngineRedundancy::getResultAngle()
{
  CHECK_INITIALIZED(TimedDoubleSeq());
  
  
  TimedDoubleSeq q;
  q.data.length(23);
  

  for (unsigned int i = 0; i < q.data.length(); ++i) {
    q.data[i] = ref_model->virtualJoint(i)->q();
  }
  
  return q;
}




double EngineRedundancy::getResultElbowAngle(RightLeft type)
{
  CHECK_INITIALIZED(double ());
  

  StandardModelPart* part_arm;
  Link* wrist;
  if(type == RIGHT) {
    part_arm = ref_model->part("RARM");
    wrist  = ref_model->virtualJoint("R_END")->endLink(); 
  }
  else if(type == LEFT) {
    part_arm = ref_model->part("LARM");
    wrist  = ref_model->virtualJoint("L_END")->endLink(); 
  }
  else {
    cout << "error : [Engine] [getResultElbowAngle] : No such arm" << endl;
    return 0;
  }
  
  
  if( part_arm->virtualJoint(3)->empty() ) {
    return 0;
  }

  
  if( part_arm->numRealJoints() < 7 ) {
    return 0;
  }
  
  
  // 肘角度の計算
  Link* shoulder = part_arm->rootVirtualJoint()->rootLink();
  Link* elbow    = part_arm->virtualJoint(3)->endLink();
 


  vector3 w(wrist->p - shoulder->p);   // 肩->手首ベクトル
  vector3 hw(tvmet::normalize(w));     // wのnormalized vector(hはhatの意味)
  vector3 e(elbow->p - shoulder->p);   // 肩->肘ベクトル
  vector3 hv(0, 0, 1);                 // 基準面作成用ベクトル
  vector3 d(hw * (tvmet::dot(hw, e))); // wが作る直線と肘からおろした垂線上にある点
  vector3 p(e - d);                    // 点dから肘へのベクトル (p = e - d)


  double at0,at1;
  at0 = tvmet::dot(hw, tvmet::cross(hv, p));
  at1 = tvmet::dot(hv, p);
  double phi = atan2(at0, at1);  // 肘角度


  return phi;
}





//------------------------------------------------------------------
//  for service port
//------------------------------------------------------------------

bool EngineRedundancy::resetOriginalFrame()
{
  CHECK_INITIALIZED(false);

  for(int i = 0; i < 3; i++) {
    cur_model->virtualJoint(i)->q(0.0);
    ref_model->virtualJoint(i)->q(0.0);
  }
  
  cur_model->calcForwardKinematics();
  ref_model->calcForwardKinematics();
  
  return true;
}



bool EngineRedundancy::selectRedundancy(const RedundantAxesMask& redundancy)
{
  ////////  move  ////////
  if(ref_model->part("MOVE")->numRealJoints() > 0) {
    use_move = false;
    select_move = dzeromatrix(3,3);

    if(redundancy.vehicleX != 0 && !ref_model->virtualJoint("BASE_X")->empty()) {
      select_move(0,0) = 1.0;
      use_move = true;
    }
    if(redundancy.vehicleY != 0 && !ref_model->virtualJoint("BASE_Y")->empty()) {
      select_move(1,1) = 1.0;
      use_move = true;
    }
    if(redundancy.vehicleTheta != 0 && !ref_model->virtualJoint("BASE_T")->empty()) {
      select_move(2,2) = 1.0;
      use_move = true;
    }
  }


  ////////  waist  ////////
  if(ref_model->part("WAIST")->numRealJoints() > 0) {
    use_waist = false;
    select_waist = dzeromatrix(3,3);

    if(redundancy.lumbarRz != 0 && !ref_model->virtualJoint("WAIST1")->empty()) {
      select_waist(0,0) = 1.0;
      use_waist = true;
    }
    if(redundancy.lumbarRy != 0 && !ref_model->virtualJoint("WAIST2")->empty()) {
      select_waist(1,1) = 1.0;
      use_waist = true;
    }
    if(redundancy.lumbarRx != 0 && !ref_model->virtualJoint("WAIST3")->empty()) {
      select_waist(2,2) = 1.0;
      use_waist = true;
    }
  }


  ////////  elbow  ////////
  use_elbow_r = redundancy.rightElbow;
  use_elbow_l = redundancy.leftElbow;

  return true;
}



bool EngineRedundancy::setControlPointOffset(RightLeft type, const HgMatrix offset)
{
  CHECK_INITIALIZED(false);


  vector3  offset_p = vector3(offset[0][3] / 1000.0, offset[1][3] / 1000.0, offset[2][3] / 1000.0);
  matrix33 offset_R;
  for(int i = 0; i < 3; i++)
    for(int j = 0; j < 3; j++)
      offset_R(i,j) = offset[i][j];


  if(type == RIGHT) {
    cur_model->virtualJoint("R_END")->endLink()->b  = offset_p;
    cur_model->virtualJoint("R_END")->endLink()->Rs = offset_R;

    ref_model->virtualJoint("R_END")->endLink()->b  = offset_p;
    ref_model->virtualJoint("R_END")->endLink()->Rs = offset_R;
  }
  else if(type == LEFT) {
    cur_model->virtualJoint("L_END")->endLink()->b  = offset_p;
    cur_model->virtualJoint("L_END")->endLink()->Rs = offset_R;
    
    ref_model->virtualJoint("L_END")->endLink()->b  = offset_p;
    ref_model->virtualJoint("L_END")->endLink()->Rs = offset_R;
  }
  else
    return false;


  return true;
}






////////////////////////////////////////////////////////////////////
//
//    private function
//
////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------
//  for status port
//------------------------------------------------------------------

extern "C"{//もともとFortranの関数->C言語->C++に変換

  int dgesvd_(char *jobu, char *jobvt, long int *m, long int  *n, double *a,
	      long int  *lda, double *s, double *u, long int *ldu,
	      double *vt, long int  *ldvt, double *work, long int  *lwork, long int *info);
}

double EngineRedundancy::SVD(const dmatrix &J)
{  
  char jobu = 'N';
  char jobvt = 'N';

  long int m = (long int)J.size1(), n = (long int)J.size2(), min_mn = min(m, n);
  long int lda = m, ldu = m, ldvt=1;
  long int lwork = max(3*min_mn+max(m,n), 5*min_mn);
  long int info;

  double s[min_mn], u[m*min_mn], vt[1], work[lwork];

  double A[m*n];
  
  for(unsigned int i=0;i<n;i++){
    for(unsigned int j=0;j<m;j++){
      A[i*m + j] = J(j,i);
    }
  }
  //SVD本体，特異値が降順にs[0]から入っていく
  dgesvd_(&jobu, &jobvt, &m, &n, A, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, &info);


  if(info < 0){
    cout << "error : [Engine] [SVD] : No."<< -info  << " calcuration failure " << endl;
    return -1;
  }
  else if(info > 0){
    cout << "error : [Engine] [SVD] : can not convergence the calcuration " << endl;
    return -2;
  }  

  return s[min_mn -1] / s[0];//最小特異値 / 最大特異値
} 



//------------------------------------------------------------------
//  for inverse kinematics
//------------------------------------------------------------------

dvector EngineRedundancy::calcDeltaP(const vector3 &p_ref_in, const matrix33 &R_ref_in, RightLeft type)
{
  Link* end_link;
  if(type == RIGHT) {
    end_link = ref_model->virtualJoint("R_END")->endLink();
  }
  else {
    end_link = ref_model->virtualJoint("L_END")->endLink();
  }


  ////////  calcurate generalized motion velocity vector(u_ref)  ////////
  // position (v)
  vector3 v(p_ref_in - end_link->p);


  // attitude (omega)
  matrix33 R_cur = end_link->R;
  matrix33 R_ref(R_ref_in * trans(end_link->Rs));
  matrix33 dR = R_ref.M_sub_eq(R_cur);

  
  // avoid aliasing of tvmet library
  matrix33 R_cur_T;
  R_cur_T = tvmet::trans(R_cur);

  
  // calcurate angular velocity matrix S
  // S = dR*R_T
  dmatrix dR_dmat(3,3);
  dmatrix R_cur_T_dmat(3,3);
  for(int a=0;a<3;a++){
    for(int b=0;b<3;b++){
      dR_dmat(a,b)=dR(a,b);
      R_cur_T_dmat(a,b)=R_cur_T(a,b);
    }
  }
  dmatrix S_dmat = ublas::prod(dR_dmat,R_cur_T_dmat);

  
  // get angular velocity vector (omega)
  vector3 omega;
  omega(0) = S_dmat(2,1);
  omega(1) = S_dmat(0,2);
  omega(2) = S_dmat(1,0);
  
  
  // make generalized task velocity vector
  dvector u_ref(6);
  for (unsigned int i=0; i<3; i++) {
    u_ref(i) = v(i);
    u_ref(i+3) = omega(i);
  }
  
  return u_ref;
}


// SC擬似逆行列による冗長マニピュレータの制御
// 090217修正版　判別方法の変更
bool EngineRedundancy::SCmethod(dvector &d_theta, const dmatrix &Jac, const dvector6 &d_x_ref) const
{
  const double d_theta_limit = limmiter;
  dmatrix J_T     = ublas::trans(Jac);
  dmatrix JJT     = ublas::prod(Jac, J_T);
  dmatrix adj_JJT = calcAdjMatrix(JJT);        //余因子行列を計算
  dmatrix J_tmp   = ublas::prod(J_T, adj_JJT);
  double det_JJT  = det(JJT);
  double b;

  // SC法を用いるかどうかを行列式の大きさで判別
  if(fabs(det_JJT) < 0.001){
    // SC法を適応する
    if(det_JJT > 0){
      b = 1000.0;
    }
    else{
      b = -1000.0;
    }
  }
  else{
    // 通常モードで処理するため、falseで処理を抜ける
    return false;
  }
  // SC法によるd_thetaの計算を開始する
  
  dvector d_theta_tmp = b * ublas::prod(J_tmp, d_x_ref);

  // 関節の変化量のリミットを越えないようにする
  double d_theta_tmp_max = fabs(d_theta_tmp[0]);
  for(unsigned int i = 1; i < d_theta_tmp.size(); i++){
    if(fabs(d_theta_tmp[i]) > d_theta_tmp_max)
      d_theta_tmp_max = fabs(d_theta_tmp[i]);
  }

  /*
   * 各関節の中の最大変化量d_theta_tmp_maxが
   * リミット(d_theta_limit)より大きかったら，
   * 関節の変化量の最大値 で 各関節の変化量を正規化し，
   * 関節の最大変化量がリミットになるようにする．
   */
  if(d_theta_tmp_max > d_theta_limit){
      d_theta = d_theta_limit * (d_theta_tmp / d_theta_tmp_max);
  }
  else{
    d_theta = d_theta_tmp;
  }
  //ここまでで，SC法による関節角度の変化量 d_theta の計算は終了

  // SC法適応(true)で終わる
  return true;
}




bool EngineRedundancy::calcMinimizeNorm(const vector3 p_ref, const matrix33 R_ref, RightLeft type, bool withWaist, bool withMove)
{
  StandardModelPart* part_arm;
  VirtualJoint*      end_virtual_joint;

  if(type == RIGHT) {
    part_arm = ref_model->part("RARM");
    end_virtual_joint = ref_model->virtualJoint("R_END");
  }
  else if(type == LEFT) {
    part_arm = ref_model->part("LARM");
    end_virtual_joint = ref_model->virtualJoint("L_END");
  }
  else {
    cout << "error : [Engine] [calcMinimizeNorm] : No such arm" << endl;
    return false;
  }


  if(part_arm->numRealJoints() < 6) {
    //return false;  // 計算ミス扱いにされるのでtrueを返す(6軸未満用)
    return true;
  }



  ////////  calc Jacobian  ////////
  dmatrix Jac = part_arm->Jacobian(end_virtual_joint);
  dmatrix select_matrix = didentity(part_arm->numVirtualJoints());
  if(withWaist) {
    dmatrix Jac_waist = ublas::prod(ref_model->part("WAIST")->Jacobian(end_virtual_joint), select_waist);
    Jac = addToMat(Jac_waist, Jac, Right);
    select_matrix = diag(select_waist, select_matrix);
  }
  if(withMove) {
    dmatrix Jac_move = ublas::prod(ref_model->part("MOVE")->Jacobian(end_virtual_joint), select_move);
    Jac = addToMat(Jac_move, Jac, Right);
    select_matrix = diag(select_move, select_matrix);
  }


  // for status port
  condition_number = SVD(Jac);



  dmatrix Jac_Pinv;
  if(calcPseudoInverse(Jac, Jac_Pinv)) {
    cout << "error : [Engine] [minimize] : calc pseudo inverse" << endl;
    return false;
  }


 
  ////////  calc delta p  ////////
  dvector u_ref = calcDeltaP(p_ref, R_ref, type);



  ////////  calc virtual angle  ////////
  dvector delta_q_ref, delta_q_ref_virtual;
  delta_q_ref = ublas::prod(Jac_Pinv, u_ref);
  delta_q_ref_virtual = ublas::prod(select_matrix, delta_q_ref);
 


  ////////  update reference model  ////////
  unsigned int i = 0;  
  if(withMove) {
    for(int j = 0; j < ref_model->part("MOVE")->numVirtualJoints(); j++) {
      double q_tmp = ref_model->part("MOVE")->virtualJoint(j)->q();
      ref_model->part("MOVE")->virtualJoint(j)->q( q_tmp + delta_q_ref_virtual(i) );
      i++;
    }
  }
  if(withWaist) {
    for(int j = 0; j < ref_model->part("WAIST")->numVirtualJoints(); j++) {
      double q_tmp = ref_model->part("WAIST")->virtualJoint(j)->q();
      ref_model->part("WAIST")->virtualJoint(j)->q( q_tmp + delta_q_ref_virtual(i) );
      i++;
    }
  }
  for(int j = 0; j < part_arm->numVirtualJoints(); j++) {
    double q_tmp = part_arm->virtualJoint(j)->q();
    part_arm->virtualJoint(j)->q( q_tmp + delta_q_ref_virtual(i) );
    i++;
  }
  ref_model->calcForwardKinematics();


  return checkResults(type, false);
}



bool EngineRedundancy::calcElbowAngle(const vector3 p_ref, const matrix33 R_ref, const double elbow_ref, RightLeft type)
{
  StandardModelPart* part_arm;
  VirtualJoint*      end_virtual_joint;
  
  if(type == RIGHT) {
    part_arm = ref_model->part("RARM");
    end_virtual_joint = ref_model->virtualJoint("R_END");
  }
  else if(type == LEFT) {
    part_arm = ref_model->part("LARM");
    end_virtual_joint = ref_model->virtualJoint("L_END");
  }
  else {
    cout << "error : [Engine] [calcMinimizeNorm] : No such arm" << endl;
    return false;
  }


  // not exist elbow
  if( part_arm->virtualJoint(3)->empty() ) {
    return true;
  }


  // elbow-angle is 7 DOF solution
  if( part_arm->numRealJoints() < 7 ) {
    // 計算ミス扱いにされるのでtrueを返す(単腕時の対策)
    return true;
  }
  


  ////////  elbow angle method  ////////
  Link* shoulder = part_arm->rootVirtualJoint()->rootLink();
  Link* elbow    = part_arm->virtualJoint(3)->endLink();
  Link* wrist    = end_virtual_joint->endLink();

  dmatrix Jw_whole = part_arm->Jacobian(end_virtual_joint);         // 6 x 7 matrix
  dmatrix Je_whole = part_arm->Jacobian(part_arm->virtualJoint(3)); // 6 x 7 matrix *** 3列まで有効，5-7列は使ってはいけない
  dzeromatrix temp(3, 7);
  dmatrix Jw(temp), Je(temp); // 3 x 7 zero matrix

  // 速度(vx,vy,vz)に関する項を抜き出す
  for (unsigned int i = 0; i < 3; i++) {
    for (unsigned int j = 0; j < 7; j++) {
      Jw(i,j) = Jw_whole(i,j);
    }
    for (unsigned int j = 0; j < 3; j++) {
      Je(i,j) = Je_whole(i,j);
    }
  }


  // --- tvmet version ---
  vector3 w(wrist->p - shoulder->p);     // 肩->手首ベクトル
  vector3 e(elbow->p - shoulder->p);     // 肩->肘ベクトル
  vector3 hw(tvmet::normalize(w));       // wのnormalized vector(hはhatの意味)
  vector3 he(tvmet::normalize(e));       // e     〃
  vector3 hv(0, 0, 1);                   // 基準面作成用ベクトル
  vector3 d(hw * (tvmet::dot(hw, e)));   // wが作る直線と肘からおろした垂線上にある点
  vector3 p(e - d);                      // 点dから肘へのベクトル (p = e - d)
  vector3 l = tvmet::cross(tvmet::cross(w, hv), w);
  vector3 hp(tvmet::normalize(p));
  vector3 hl(tvmet::normalize(l));


  // for "v" and "p" are same case. (self-motion)
  // but this code doesn't solve problem...
  double v_eq_p_epsilon = 1E-2;
  double elbow_cur;
  elbow_cur = atan2( tvmet::dot(hw, tvmet::cross(hv, p)) , tvmet::dot(hv, p) ); //  式2.66
  /*
  if( fabs(tvmet::dot(hv, p)) <= v_eq_p_epsilon) {
    // this code doesn't enough to solve self-mottion.
    elbow_cur = elbow_ref;
  }
  else {
    elbow_cur = atan2( tvmet::dot(hw, tvmet::cross(hv, p)) , tvmet::dot(hv, p) ); //  式2.66
  }
  */

  // ここから式2.68の計算  
  vector3 Je_coefficient_tvmet, Jw_coefficient_tvmet;
  Je_coefficient_tvmet = tvmet::cross(hw, hp) / tvmet::norm2(p);

  vector3 Jw_co_A, Jw_co_B;
  Jw_co_A = ( tvmet::dot(hv, w) / tvmet::norm2(l) ) * tvmet::cross(hw, hl);
  Jw_co_B = ( tvmet::dot(hw, he) / tvmet::norm2(p) ) * tvmet::cross(hw, hp);
  Jw_coefficient_tvmet = Jw_co_A.M_sub_eq(Jw_co_B);
  

  // --- ublas version ---
  dvector Je_coefficient_boost(3), Jw_coefficient_boost(3);  
  for(unsigned int i = 0; i < 3; i++) {
    Je_coefficient_boost(i) =  Je_coefficient_tvmet(i);
    Jw_coefficient_boost(i) =  Jw_coefficient_tvmet(i);
  }

  dvector J_psi(7);
  dvector J_psi_e = ublas::prod(Je_coefficient_boost, Je); // 1 x 7
  dvector J_psi_w = ublas::prod(Jw_coefficient_boost, Jw); // 1 x 7
  J_psi = J_psi_e + J_psi_w; 
  // ここまで式2.68


  // ここから式2.70の計算
  // calc elbow Jacobian
  dmatrix Jac_arm = part_arm->Jacobian(end_virtual_joint);  // 6 x 7
  dmatrix Jac(7,7);
  for(unsigned int i = 0; i < 7; i++) {
    for(unsigned int j = 0; j < 6; j++) {
      Jac(j,i) = Jac_arm(j,i);
    }
    Jac(6,i) = J_psi(i);
  }
  // ここまで式2.70



  // for status port
  condition_number = SVD(Jac);




  ////////  calc delta p  ////////
  dvector u_ref = calcDeltaP(p_ref, R_ref, type);
  dvector u_ref_elbow(7);
  for (unsigned int i = 0; i < 6; i++) {
    u_ref_elbow(i) = u_ref(i);
  }
  u_ref_elbow(6) = (elbow_ref - elbow_cur);


  if( u_ref_elbow(6) > M_PI )
    u_ref_elbow(6) -= ( 2 * M_PI );
  else if( u_ref_elbow(6) < -M_PI )
    u_ref_elbow(6) += ( 2 * M_PI );


  ////////  calc virtual angle  ////////
  dmatrix Jac_inv = inverse(Jac);
  dvector delta_theta = ublas::prod(Jac_inv, u_ref_elbow);
  

  ////////  update reference model  ////////
  for(int i = 0; i < part_arm->numVirtualJoints(); i++) {
    double q_tmp = part_arm->virtualJoint(i)->q();
    part_arm->virtualJoint(i)->q( q_tmp + delta_theta(i) );
  }
  ref_model->calcForwardKinematics();


  return checkResults(type, true);
}


bool EngineRedundancy::calcMinimizeNormDual(const vector3 p_ref_r, const matrix33 R_ref_r, const vector3 p_ref_l, const matrix33 R_ref_l, bool withWaist, bool withMove)
{
  // 単腕用
  if(ref_model->part("RARM")->numRealJoints() < 6) {
    return calcMinimizeNorm(p_ref_l, R_ref_l, LEFT, withWaist, withMove);
  }
  else if(ref_model->part("LARM")->numRealJoints() < 6) {
    return calcMinimizeNorm(p_ref_r, R_ref_r, RIGHT, withWaist, withMove);
  }


  //            |              |               |          |          | 
  //            | (Jac_r_move) | (Jac_r_waist) | Jac_rarm |   zero   | 
  //  Jac_all = |----------------------------------------------------|  ... 12 x 20
  //            | (Jac_l_move) | (Jac_l_waist) |   zero   | Jac_larm |
  //            |              |               |          |          |


  ////////  calc Jacobian  ////////
  VirtualJoint* r_end = ref_model->virtualJoint("R_END");
  VirtualJoint* l_end = ref_model->virtualJoint("L_END");

  dmatrix Jac_rarm = ref_model->part("RARM")->Jacobian(r_end);
  dmatrix Jac_larm = ref_model->part("LARM")->Jacobian(l_end);

  dmatrix Jac_all = diag(Jac_rarm, Jac_larm);
  dmatrix select_matrix = didentity(14);

  if(withWaist) {
    dmatrix Jac_r_waist = ublas::prod( ref_model->part("WAIST")->Jacobian(r_end), select_waist );
    dmatrix Jac_l_waist = ublas::prod( ref_model->part("WAIST")->Jacobian(l_end), select_waist );
    
    dmatrix Jac_waist = addToMat(Jac_r_waist, Jac_l_waist, Lower);
    Jac_all = addToMat(Jac_waist, Jac_all, Right);

    select_matrix = diag(select_waist, select_matrix);
  }
  if(withMove) {
    dmatrix Jac_r_move = ublas::prod( ref_model->part("MOVE")->Jacobian(r_end), select_move );
    dmatrix Jac_l_move = ublas::prod( ref_model->part("MOVE")->Jacobian(l_end), select_move );
    
    dmatrix Jac_move = addToMat(Jac_r_move, Jac_l_move, Lower);
    Jac_all = addToMat(Jac_move, Jac_all, Right);

    select_matrix = diag(select_move, select_matrix);
  }



  // for status port
  condition_number = SVD(Jac_all);



  dmatrix Jac_Pinv;
  if(calcPseudoInverse(Jac_all, Jac_Pinv)) {
    cout << "error : [Engine] [minimize] : calc pseudo inverse" << endl;
    return false;
  }



  ////////  calc delta p  ////////
  dvector u_r = calcDeltaP(p_ref_r, R_ref_r, RIGHT);
  dvector u_l = calcDeltaP(p_ref_l, R_ref_l, LEFT);

  dvector u_ref(12);
  for(int i = 0; i < 6; i++) {
    u_ref(i)   = u_r(i);
    u_ref(i+6) = u_l(i);
  }


  ////////  calc virtual angle  ////////
  dvector delta_q_ref, delta_q_ref_virtual;
  delta_q_ref = ublas::prod(Jac_Pinv, u_ref);
  delta_q_ref_virtual = ublas::prod(select_matrix, delta_q_ref);
  

 
  ////////  update reference model  ////////
  unsigned int i = 0;  
  if(withMove) {
    for(int j = 0; j < ref_model->part("MOVE")->numVirtualJoints(); j++) {
      double q_tmp = ref_model->part("MOVE")->virtualJoint(j)->q();
      ref_model->part("MOVE")->virtualJoint(j)->q( q_tmp + delta_q_ref_virtual(i) );
      i++;
    }
  }
  if(withWaist) {
    for(int j = 0; j < ref_model->part("WAIST")->numVirtualJoints(); j++) {
      double q_tmp = ref_model->part("WAIST")->virtualJoint(j)->q();
      ref_model->part("WAIST")->virtualJoint(j)->q( q_tmp + delta_q_ref_virtual(i) );
      i++;
    }
  }
  for(int j = 0; j < ref_model->part("RARM")->numVirtualJoints(); j++) {
    double q_tmp = ref_model->part("RARM")->virtualJoint(j)->q();
    ref_model->part("RARM")->virtualJoint(j)->q( q_tmp + delta_q_ref_virtual(i) );
    i++;
  }
  for(int j = 0; j < ref_model->part("LARM")->numVirtualJoints(); j++) {
    double q_tmp = ref_model->part("LARM")->virtualJoint(j)->q();
    ref_model->part("LARM")->virtualJoint(j)->q( q_tmp + delta_q_ref_virtual(i) );
    i++;
  }
  ref_model->calcForwardKinematics();



  return checkResults(DUAL, false);
}




bool EngineRedundancy::checkResults(RightLeft type, bool withElbow, bool withComment)
{
  double maxIkErrorPos   = 1.0e-4;
  double maxIkErrorElbow = 3.0e-4;
  bool   calcSuccess = true;


  // check position error
  if(type == RIGHT || type == DUAL) {
    dvector dp   = calcDeltaP(dst_pos[RIGHT], dst_rot[RIGHT], RIGHT);
    double error = 0;

    for(unsigned int i = 0; i < dp.size(); i++) {
      error += ( dp(i) * dp(i) );
    }
    if(error > maxIkErrorPos) {
      calcSuccess = false;
    }

    // for debug
    if(withComment) {
      cout << "    POSITION ERROR RIGHT  :  " << error;
    }
  }
  if(type == LEFT || type == DUAL) {
    dvector dp   = calcDeltaP(dst_pos[LEFT], dst_rot[LEFT], LEFT);
    double error = 0;
    
    for(unsigned int i = 0; i < dp.size(); i++) {
      error += ( dp(i) * dp(i) );
    }
    if(error > maxIkErrorPos) {
      calcSuccess = false;
    }

    // for debug
    if(withComment) {
      cout << "    POSITION ERROR LEFT  :  " << error;
    }
  }



  // check elbow-angle error
  if(withElbow) {
    if(type == RIGHT || type == DUAL) {
      double error = ( dst_redun[6] - getResultElbowAngle(RIGHT) );

      if( error > M_PI )
	error -= ( 2 * M_PI );
      else if( error < -M_PI )
	error += ( 2 * M_PI );
      
      error = (error * error);
      if(error > maxIkErrorElbow) {
	calcSuccess = false;
      }

      // for debug
      if(withComment) {
	cout << "    ELBOW ERROR RIGHT  :  " << error;
      }
    }
    
    if(type == LEFT || type == DUAL) {
      double error = ( dst_redun[7] - getResultElbowAngle(LEFT) );

      if( error > M_PI )
	error -= ( 2 * M_PI );
      else if( error < -M_PI )
	error += ( 2 * M_PI );
      
      error = (error * error);
      if(error > maxIkErrorElbow) {
	calcSuccess = false;
      }

      // for debug
      if(withComment) {
	cout << "    ELBOW ERROR RIGHT  :  " << error;
      }
    }
  }
  // for debug
  if(withComment) {
    cout << endl;
  }

  return calcSuccess;  
}







//------------------------------------------------------------------
//  matrix tools
//------------------------------------------------------------------

dvector3 EngineRedundancy::cross_prod(dvector3 &a, dvector3 &b)
{
  dvector3 Result;
  Result(0) = a(1)*b(2) - a(2)*b(1);
  Result(1) = a(2)*b(0) - a(0)*b(2);
  Result(2) = a(0)*b(1) - a(1)*b(0);
  return Result;
}



dmatrix EngineRedundancy::diag(const dmatrix &in1, const dmatrix &in2)
{
  int size1 = (int)in1.size1() + (int)in2.size1();
  int size2 = (int)in1.size2() + (int)in2.size2();
  
  
  dmatrix out(size1, size2);
  dzeromatrix zero(size1, size2);
  out = zero;
  
  
  for(int i = 0; i < (int)in1.size1(); i++)
    for(int j = 0; j < (int)in1.size2(); j++)
      out(i,j) = in1(i,j);
  
  
  int shift1 = (int)in1.size1();
  int shift2 = (int)in1.size2();
  for(int i = 0; i < (int)in2.size1(); i++)
    for(int j = 0; j < (int)in2.size2(); j++)
      out(i+shift1,j+shift2) = in2(i,j);
  
  return out;
}



dmatrix EngineRedundancy::addToMat(const dmatrix &in1, const dmatrix &in2, const Direction dir)
{
  dmatrix out;

  if(in1.size1() == 0 || in1.size2() == 0)
    return in2;
  else if(in2.size1() == 0 || in2.size2() == 0)
    return in1;
  

  if(dir == Right || dir == Left)
    if(in1.size1() != in2.size1())
      return out;

  if(dir == Lower || dir == Upper)
    if(in1.size2() != in2.size2())
      return out;


  unsigned int row = 0, col = 0;
  unsigned int shift1 = 0, shift2 = 0;

  if(dir == Lower) {
    row = in1.size1() + in2.size1();
    col = in1.size2();

    shift1 = in1.size1();
    shift2 = 0;
  }
  else if(dir == Right) {
    row = in1.size1();
    col = in1.size2() + in2.size2();

    shift1 = 0;
    shift2 = in1.size2();
  }
  else if(dir == Upper){
    return addToMat(in2, in1, Lower);
  }
  else if(dir == Left){
    return addToMat(in2, in1, Right);
  }

  out.resize(row, col);
  for(unsigned int i = 0; i < in1.size1(); i++) {
    for(unsigned int j = 0; j < in1.size2(); j++) {
      out(i,j) = in1(i,j);
    }
  }
  for(unsigned int i = 0; i < in2.size1(); i++) {
    for(unsigned int j = 0; j < in2.size2(); j++) {
      out(i+shift1,j+shift2) = in2(i,j);
    }
  }
  
  return out;
}
















//-------------------------------------------------------------------------------------
//   for test
//-------------------------------------------------------------------------------------

void EngineRedundancy::test()
{
  for(int i = 0; i < ref_model->numParts(); i++) {
    cout << "  " << ref_model->part(i)->name << endl;;

    for(int j = 0; j < ref_model->part(i)->numVirtualJoints(); j++) {
      cout << "    " << ref_model->part(i)->virtualJoint(j)->name << "  :";

      for(int k = 0; k < ref_model->part(i)->virtualJoint(j)->numLinks(); k++) {
	cout << "  " << ref_model->part(i)->virtualJoint(j)->link(k)->name;
      }
      cout << endl;
    }
  }


  cout << "\n\n" << endl;
  JointPathPtr joints = ref_model->body->getJointPath(ref_model->body->rootLink(), ref_model->body->link("R_END"));

  StandardModelPart* part_arm;
  VirtualJoint*      end_virtual_joint;
  part_arm = ref_model->part("RARM");
  end_virtual_joint = ref_model->virtualJoint("R_END");
 


  cout << joints->Jacobian() << endl;
  cout << ref_model->part("RARM")->Jacobian(end_virtual_joint) << endl;

  dmatrix Jac;
  Jac = ref_model->part("RARM")->Jacobian(end_virtual_joint);
  Jac = addToMat(ref_model->part("WAIST")->Jacobian(end_virtual_joint), Jac, Right);
  cout << Jac << endl;

}



#endif
