#include "ModelCreator.h"
#include "JointName.h"

#include <sys/param.h>

using namespace OpenHRP;
using namespace std;
using namespace tohoku;


bool tohoku::createRealModelFromJointNameMap(BodyPtr &body, JointName* jointname)
{
  //----------------------------------------------------------------
  // load base model
  //----------------------------------------------------------------
  char current_dir[MAXPATHLEN];
  char *p = NULL;
  
  p = getcwd(current_dir, MAXPATHLEN);
  if (p == NULL) {
    perror("getcwd");
    return false;
  }
  
  stringstream url;
  url << "file://";
  url << current_dir;
  url << "/";
  url << jointname->getModelFile();

  
  BodyPtr base = new Body();
  if ( !loadBodyFromModelLoader(base, url.str().c_str(), 0, NULL) ) {
    std::cerr << "failed to load model[" << url.str().c_str() << "]" << std::endl;
    return false;
  }



  //----------------------------------------------------------------
  // mapping
  //----------------------------------------------------------------
  body = new Body();

  ////////  create root link  ////////
  Link* root = new Link();
  root->jointType = Link::FREE_JOINT;
  root->name = "BASE_JOINT";
  root->a    = vector3(0,0,0);
  root->b    = vector3(0,0,0);
  //root->p    = vector3(0,0,0);
  root->p    = base->rootLink()->p;
  root->b    = vector3(0,0,0);
  root->R    = tvmet::identity<matrix33>();
  root->R    = base->rootLink()->R;
  //root->Rs   = tvmet::identity<matrix33>();
  
  body->setRootLink(root);
  body->updateLinkTree();  



  ////////  create part joints  ////////
  vector< vector<Link*> > parts;
  int jointId = 0;

  // MOVE
  vector<Link*> move_joints = createMoveJointsInPart(base, jointname, jointId);
  parts.push_back(move_joints);
  jointId += (int)move_joints.size();

  // WAIST, RARM, LARM, NECK
  for(int i = 1; i < jointname->numParts()-1; i++) {
    vector<Link*> joints = createJointsInPart(base, jointname, i, jointId);
    parts.push_back(joints);
    jointId += (int)joints.size();
  }



  ////////  add part joints to body  ////////
  int num_parts = parts.size();
  int num_lower = 2;
  //int num_upper = num_parts - num_lower;

  Link* parent = root;

  // lower
  for(int i = 0; i < num_lower; i++) {
    for(unsigned int j = 0; j < parts[i].size(); j++) {
      parent->addChild(parts[i][j]);
      parent = parts[i][j];
    }
  }
  Link* lower_end = parent;  

  // upper
  for(int i = num_parts-1; i >= num_lower; i--) {
    for(unsigned int j = 0; j < parts[i].size(); j++) {
      parent->addChild(parts[i][j]);
      parent = parts[i][j];
    }
    parent = lower_end;
  }
  

  
  ////////  update model  //////
  body->updateLinkTree();  
  for(int i = 0; i < body->numLinks(); i++) {
    body->link(i)->q = 0.0;
  }
  body->calcForwardKinematics(); 



  return true;
}



bool tohoku::createStandardModelFromJointNameMap(StandardModel* stdModel, JointName* jointname)
{
  // パーツリスト取得
  vector<string> part_names = jointname->getVirtualJointPartNames();  
  for(int i = 0; i < jointname->numParts(); i++) {

    // パーツ作成
    StandardModelPart* part = new StandardModelPart(part_names[i]);

    // パーツ中の仮想ジョイントリスト取得
    vector<string> joint_names_in_part = jointname->getVirtualJointNamesInPart(part_names[i]);
    if(part_names[i] == "EXTRA") {
      joint_names_in_part.push_back("R_END");
      joint_names_in_part.push_back("L_END");
    }

    for(unsigned int j = 0; j < joint_names_in_part.size(); j++) {

      // 仮想ジョイント作成
      VirtualJoint* virtual_joint = new VirtualJoint(joint_names_in_part[j]);

      // 仮想ジョイント中の実ジョイントリスト取得
      vector<string> real_names  = jointname->getNameReal(joint_names_in_part[j]);
      vector<double> real_ratios = jointname->getRatio(joint_names_in_part[j]);
      if(joint_names_in_part[j] == "R_END") {
	real_names.push_back("R_END");   real_ratios.push_back(1.0);
      }
      if(joint_names_in_part[j] == "L_END") {
	real_names.push_back("L_END");   real_ratios.push_back(1.0);
      }

      // 仮想ジョイントに実ジョイント追加
      double all_ratio = 0.0;
      for(unsigned int k = 0; k < real_names.size(); k++) {
	all_ratio += real_ratios[k];
      }
      if(fabs(all_ratio) < 1.0e-6) {
	all_ratio = 1.0;
      }
      for(unsigned int k = 0; k < real_names.size(); k++) {
	Link* real_link = stdModel->body->link(real_names[k]);
	virtual_joint->addMember(real_link, real_ratios[k] / all_ratio);
      }

      // パーツに仮想ジョイント追加
      part->addVirtualJoint(virtual_joint);      
    }

    // 標準モデルにパーツ追加
    stdModel->addPart(part);
  }
  
  stdModel->updateStandardModel();

  return true;
}





vector<Link*> tohoku::createJointsInPart(BodyPtr &body, JointName *jointname, int index, int jointId)
{
  vector<Link*> joints;
  joints.clear();

  vector<string> virtual_joint_names = jointname->getVirtualJointNamesInPart(index);
  for(unsigned int i = 0; i < virtual_joint_names.size(); i++) {

    vector<string> real_joint_names = jointname->getNameReal(virtual_joint_names[i]);
    if( !real_joint_names.empty() )
      for(unsigned int j = 0; j < real_joint_names.size(); j++) {
	
	body->link(real_joint_names[j])->parent  = 0;
	body->link(real_joint_names[j])->sibling = 0;
	body->link(real_joint_names[j])->child   = 0;
	
	Link* link = new Link();
	link->copy(*body->link(real_joint_names[j]));
	
	link->q       = 0.0;
	link->jointId = jointId;
	jointId++;
	
	joints.push_back(link);
      }  
  }


  if(jointname->getVirtualJointPartNames()[index] == "RARM") {
    Link* r_end = new Link();
    r_end->name = "R_END";
    r_end->jointType = Link::FIXED_JOINT;
    r_end->R         = tvmet::identity<matrix33>();

    vector<double> r_offset_p = jointname->getControlPointOffsetP(0);
    r_end->b         = vector3(r_offset_p[0] / 1000.0, r_offset_p[1] / 1000.0, r_offset_p[2] / 1000.0);

    vector<double> r_offset_R = jointname->getControlPointOffsetR(0);
    matrix33 tmp_r;
    int k = 0;
    for(int i = 0; i < 3; i++) {
      for(int j = 0; j < 3; j++) {
      tmp_r(i,j) = r_offset_R[k];
      k++;
      }
    }
    r_end->Rs        = tmp_r;
    
    joints.push_back(r_end);
  }

  if(jointname->getVirtualJointPartNames()[index] == "LARM") {
    Link* l_end = new Link();
    l_end->name = "L_END";
    l_end->jointType = Link::FIXED_JOINT;
    l_end->R         = tvmet::identity<matrix33>();

    vector<double> l_offset_p = jointname->getControlPointOffsetP(1);
    l_end->b         = vector3(l_offset_p[0] / 1000.0, l_offset_p[1] / 1000.0, l_offset_p[2] / 1000.0);
    
    vector<double> l_offset_R = jointname->getControlPointOffsetR(1);
    matrix33 tmp_l;
    int k = 0;
    for(int i = 0; i < 3; i++) {
      for(int j = 0; j < 3; j++) {
      tmp_l(i,j) = l_offset_R[k];
      k++;
      }
    }
    l_end->Rs        = tmp_l;
    
    joints.push_back(l_end);
  }

  return joints;
}



vector<Link*> tohoku::createMoveJointsInPart(BodyPtr &body, JointName *jointname, int jointId)
{
  vector<Link*> joints;
  joints.clear();

 
  vector<string> names_basex = jointname->getNameReal("BASE_X");
  if( !names_basex.empty() ) {
    for(unsigned int j = 0; j < names_basex.size(); j++) {
      Link* link = new Link();
      link->name      = names_basex[j];
      link->jointType = Link::SLIDE_JOINT;
      link->a         = vector3(1,0,0);
      link->d         = vector3(1,0,0);
      link->b         = vector3(0,0,0);
      link->R         = tvmet::identity<matrix33>();
      link->Rs        = tvmet::identity<matrix33>();
      link->q         = 0.0;
      link->jointId   = jointId;
      jointId++;

      joints.push_back(link);
    }
  }

  vector<string> names_basey = jointname->getNameReal("BASE_Y");
  if( !names_basey.empty() ) {
    for(unsigned int j = 0; j < names_basey.size(); j++) {
      Link* link = new Link();
      link->name      = names_basey[j];
      link->jointType = Link::SLIDE_JOINT;
      link->a         = vector3(0,1,0);
      link->d         = vector3(0,1,0);
      link->b         = vector3(0,0,0);
      link->R         = tvmet::identity<matrix33>();
      link->Rs        = tvmet::identity<matrix33>();
      link->q         = 0.0;
      link->jointId   = jointId;
      jointId++;

      joints.push_back(link);
    }
  }

  vector<string> names_baset = jointname->getNameReal("BASE_T");
  if( !names_baset.empty() ) {
    for(unsigned int j = 0; j < names_baset.size(); j++) {
      Link* link = new Link();

      link->name      = names_baset[j];
      link->jointType = Link::ROTATIONAL_JOINT;
      link->a         = vector3(1,0,0);
      link->b         = vector3(0,0,0);
      link->R         = tvmet::identity<matrix33>();
      link->Rs        = tvmet::identity<matrix33>();
      link->q         = 0.0;
      link->jointId   = jointId;
      jointId++;

      joints.push_back(link);
    }
  }

  return joints;
}
